/* $Id$

This file is part of libmspdatafile
Copyright © 2006  Mikko Rasa, Mikkosoft Productions
Distributed under the LGPL
*/

#ifndef MSP_DATAFILE_LOADER_H_
#define MSP_DATAFILE_LOADER_H_

#include <fstream>
#include <map>
#include "except.h"
#include "parser.h"
#include "statement.h"
#include "value.h"

namespace Msp {
namespace DataFile {

class Loader;
class Statement;

/**
Base class for loader actions.
*/
class LoaderAction
{
public:
	/**
	Called when a statement is to be loaded.
	*/
	virtual void execute(Loader &, const Statement &) const=0;
	virtual ~LoaderAction() { }
protected:
	LoaderAction() { }
};


/**
Loads a statement by calling a function that takes no arguments.
*/
template<typename L>
class LoaderFunc0: public LoaderAction
{
public:
	typedef void (L::*FuncType)();

	LoaderFunc0(FuncType f): func(f) { }
	void execute(Loader &l, const Statement &st) const
	{
		if(st.args.size()!=0) throw TypeError(st.get_location()+": Wrong number of arguments");
		(dynamic_cast<L &>(l).*func)();
	};
private:
	FuncType func;
};


/**
Loads a statement by calling a function that takes one argument.
*/
template<typename L, typename A0>
class LoaderFunc1: public LoaderAction
{
public:
	typedef void (L::*FuncType)(A0);

	LoaderFunc1(FuncType f): func(f) { }
	void execute(Loader &l, const Statement &st) const
	{
		if(st.args.size()!=1) throw TypeError(st.get_location()+": Wrong number of arguments");
		(dynamic_cast<L &>(l).*func)(st.args[0].get<A0>());
	}
private:
	FuncType func;
};


/**
Loads a statement by calling a function that takes an array of values.
*/
template<typename L, typename A0>
class LoaderFunc1<L, const std::vector<A0> &>: public LoaderAction
{
public:
	typedef void (L::*FuncType)(const std::vector<A0> &);

	LoaderFunc1(FuncType f): func(f) { }
	void execute(Loader &l, const Statement &st) const
	{
		std::vector<A0> values;
		values.reserve(st.args.size());
		for(ValueArray::const_iterator i=st.args.begin(); i!=st.args.end(); ++i)
			values.push_back(i->get<A0>());
		(dynamic_cast<L &>(l).*func)(values);
	}
private:
	FuncType func;
};


template<typename L, typename A0, typename A1>
class LoaderFunc2: public LoaderAction
{
public:
	typedef void (L::*FuncType)(A0, A1);

	LoaderFunc2(FuncType f): func(f) { }
	void execute(Loader &l, const Statement &st) const
	{
		if(st.args.size()!=2) throw TypeError(st.get_location()+": Wrong number of arguments");
		(dynamic_cast<L &>(l).*func)(st.args[0].get<A0>(), st.args[1].get<A1>());
	}
private:
	FuncType func;
};


template<typename L, typename A0, typename A1, typename A2>
class LoaderFunc3: public LoaderAction
{
public:
	typedef void (L::*FuncType)(A0, A1, A2);

	LoaderFunc3(FuncType f): func(f) { }
	void execute(Loader &l, const Statement &st) const
	{
		if(st.args.size()!=3) throw TypeError(st.get_location()+": Wrong number of arguments");
		(dynamic_cast<L &>(l).*func)(st.args[0].get<A0>(), st.args[1].get<A1>(), st.args[2].get<A2>());
	}
private:
	FuncType func;
};


template<typename L, typename A0, typename A1, typename A2, typename A3>
class LoaderFunc4: public LoaderAction
{
public:
	typedef void (L::*FuncType)(A0, A1, A2, A3);

	LoaderFunc4(FuncType f): func(f) { }
	void execute(Loader &l, const Statement &st) const
	{
		if(st.args.size()!=4) throw TypeError(st.get_location()+": Wrong number of arguments");
		(dynamic_cast<L &>(l).*func)(st.args[0].get<A0>(), st.args[1].get<A1>(), st.args[2].get<A2>(), st.args[3].get<A3>());
	}
private:
	FuncType func;
};


template<typename L, typename A0, typename A1, typename A2, typename A3, typename A4>
class LoaderFunc5: public LoaderAction
{
public:
	typedef void (L::*FuncType)(A0, A1, A2, A3, A4);

	LoaderFunc5(FuncType f): func(f) { }
	void execute(Loader &l, const Statement &st) const
	{
		if(st.args.size()!=5) throw TypeError(st.get_location()+": Wrong number of arguments");
		(dynamic_cast<L &>(l).*func)(st.args[0].get<A0>(), st.args[1].get<A1>(), st.args[2].get<A2>(), st.args[3].get<A3>(), st.args[4].get<A4>());
	}
private:
	FuncType func;
};


template<typename L, typename T0>
class LoadValue1: public LoaderAction
{
public:
	typedef T0 L::*Pointer0Type;

	LoadValue1(Pointer0Type p0): ptr0(p0) { }
	void execute(Loader &l, const Statement &st) const
	{
		if(st.args.size()!=1) throw TypeError(st.get_location()+": Wrong number of arguments");
		dynamic_cast<typename L::Loader &>(l).get_object().*ptr0=st.args[0].get<T0>();
	}
private:
	Pointer0Type ptr0;
};


template<typename L, typename T0>
class LoadValue1<L, T0 *>: public LoaderAction
{
public:
	typedef T0 *L::*Pointer0Type;

	LoadValue1(Pointer0Type p0): ptr0(p0) { }
	void execute(Loader &l, const Statement &st) const
	{
		if(st.args.size()!=1) throw TypeError(st.get_location()+": Wrong number of arguments");
		typename L::Loader &ldr=dynamic_cast<typename L::Loader &>(l);
		ldr.get_object().*ptr0=&ldr.get_collection().template get<T0>(st.args[0].get<std::string>());
	}
private:
	Pointer0Type ptr0;
};


template<typename L, typename T0, typename T1>
class LoadValue2: public LoaderAction
{
public:
	typedef T0 L::*Pointer0Type;
	typedef T1 L::*Pointer1Type;

	LoadValue2(Pointer0Type p0, Pointer1Type p1): ptr0(p0), ptr1(p1) { }
	void execute(Loader &l, const Statement &st) const
	{
		if(st.args.size()!=2) throw TypeError(st.get_location()+": Wrong number of arguments");
		dynamic_cast<typename L::Loader &>(l).get_object().*ptr0=st.args[0].get<T0>();
		dynamic_cast<typename L::Loader &>(l).get_object().*ptr1=st.args[1].get<T1>();
	}
private:
	Pointer0Type ptr0;
	Pointer1Type ptr1;
};


/**
Base class for data loaders.  To give loading capabilities to a class, create a
public Loader class in it, derived from this class.  Typically a loader object
contains a reference to the loaded object.  To make use of loading directly
into data members, the Loader class must have a get_object() member function,
returning that reference.  If direct loading of pointers is desired, the Loader
class must also have a get_collection() member function, returning a collection
to get pointers from.
*/
class Loader
{
public:
	/**
	Loads data from a statement.  This is normally only used by the Loader class
	itself for loading sub-items, but needs to be public so it can be accessed
	in derived objects.
	*/
	void load(const Statement &st);

	/**
	Loads statements from a parser.
	*/
	void load(Parser &p);

	virtual ~Loader();
protected:
	Loader(): cur_st(0) { }

	/**
	Adds a keyword that is loaded with a zero-argument function.
	*/
	template<typename L>
	void add(const std::string &k, void (L::*func)())
	{ add(k, new LoaderFunc0<L>(func)); }

	template<typename L, typename A0>
	void add(const std::string &k, void (L::*func)(A0))
	{ add(k, new LoaderFunc1<L, A0>(func)); }

	template<typename L, typename A0, typename A1>
	void add(const std::string &k, void (L::*func)(A0, A1))
	{ add(k, new LoaderFunc2<L, A0, A1>(func)); }

	template<typename L, typename A0, typename A1, typename A2>
	void add(const std::string &k, void (L::*func)(A0, A1, A2))
	{ add(k, new LoaderFunc3<L, A0, A1, A2>(func)); }

	template<typename L, typename A0, typename A1, typename A2, typename A3>
	void add(const std::string &k, void (L::*func)(A0, A1, A2, A3))
	{ add(k, new LoaderFunc4<L, A0, A1, A2, A3>(func)); }

	template<typename L, typename A0, typename A1, typename A2, typename A3, typename A4>
	void add(const std::string &k, void (L::*func)(A0, A1, A2, A3, A4))
	{ add(k, new LoaderFunc5<L, A0, A1, A2, A3, A4>(func)); }

	/**
	Adds a keyword that is loaded into a variable of the loaded object.
	*/
	template<typename L, typename T0>
	void add(const std::string &k, T0 L::*p0)
	{ add(k, new LoadValue1<L, T0>(p0)); }

	template<typename L, typename T0, typename T1>
	void add(const std::string &k, T0 L::*p0, T1 L::*p1)
	{ add(k, new LoadValue2<L, T0, T1>(p0, p1)); }

	/**
	Adds a keyword that is recognized but ignored.
	*/
	void add(const std::string &k)
	{ add(k, 0); }

	/**
	Loads a sub-object from the statement being processed.  The Loader class of
	the sub-object is automatically used.
	*/
	template<typename S>
	void load_sub(S &s)
	{ load_sub<typename S::Loader, S>(s); }

	/**
	Loads a sub-object with a custom Loader class.
	*/
	template<typename L, typename S>
	void load_sub(S &s)
	{
		if(!cur_st)
			throw InvalidState("load_sub called without current statement");
		L loader(s);
		loader.load(*cur_st);
	}

	template<typename S, typename T>
	void load_sub(S &s, T &p)
	{ load_sub<typename S::Loader, S, T>(s, p); }

	/**
	Loads a sub-object with a custom Loader class that takes one argument in
	addition to to object to be loaded.
	*/
	template<typename L, typename S, typename T>
	void load_sub(S &s, T &p)
	{
		if(!cur_st)
			throw InvalidState("load_sub called without current statement");
		L loader(s, p);
		loader.load(*cur_st);
	}

	/**
	Returns the source of the statement being processed.  This can be used to
	implement relative paths in include-like statements.  Note that the source
	may not necessarily be a file.
	*/
	const std::string &get_source() const
	{
		if(!cur_st)
			throw InvalidState("get_source called without current statement");
		return cur_st->source;
	}
private:
	typedef std::map<std::string, LoaderAction *> ActionMap;

	ActionMap       actions;
	const Statement *cur_st;

	void add(const std::string &, LoaderAction *);
	void load_statement(const Statement &st);
};

/**
Loads an object from a file.  The object must have a public Loader class.
*/
template<typename T>
void load(T &obj, const std::string &fn)
{
	std::ifstream in(fn.c_str());
	if(!in)
		throw Exception("Couldn't open "+fn);

	Parser parser(in, fn);
	typename T::Loader loader(obj);
	loader.load(parser);
}

template<typename T, typename U>
void load(T &obj, const std::string &fn, U arg)
{
	std::ifstream in(fn.c_str());
	if(!in)
		throw Exception("Couldn't open "+fn);

	Parser parser(in, fn);
	typename T::Loader loader(obj, arg);
	loader.load(parser);
}

} // namespace DataFile
} // namespace Msp

#endif
